import 'dart:async';
import 'dart:convert' as JSON;
import 'jsonexpression.dart';
import 'jsontoken.dart';
import 'jsontokenizer.dart';
import 'streamer.dart';

class Reader {
  JsonTokenizer tokenizer;
  Function _pipe;
  Function _trailing;
  Map<String, Function> _filters;
  Function _done;
  Function _fail;
  Reader(Streamer stream) {
    this._filters = new Map<String, Function>();
    this.tokenizer = new JsonTokenizer(stream);
    new Timer(new Duration(microseconds: 0), () {
      this._run();
    });
  }
  _run() {
    this.tokenizer.tokenize().then((item) {
      if (item == 1) {
        if (this._done != null) {
          this._done();
        }
      } else {
        try {
          this._parse();
          new Timer(new Duration(microseconds: 0), () {
            this._run();
          });
        } catch (e) {
          if (this._fail != null) {
            this._fail(e);
          }
        }
      }
    }).catchError((err) {
      if (this._fail != null) {
        this._fail(err);
      }
    });
  }

  _parse() {
    List deletes = [];
    List keys = JsonToken.queue.keys.toList();
    int len = keys.length;
    Map<int, bool> trailing = new Map<int, bool>();
    first:
    for (int i = 0; i < len; i++) {
      int key = keys[i];
      JsonToken item = JsonToken.queue[key];
      if (item == null) {
        continue;
      }
      if (item.type == JsonToken.OBJECTSTART ||
          item.type == JsonToken.ARRAYSTART) {
        int next = i + 1;
        if (next < len) {
          int _key = keys[next];
          JsonToken _item = JsonToken.queue[_key];
          trailing[_item.parnterId] = true;
        }
      }
      if (item.type == JsonToken.OBJECTEND || item.type == JsonToken.ARRAYEND) {
        int _key = keys[i - 1];
        JsonToken _item = JsonToken.queue[_key];
        trailing[_item.parnterId] = false;
      }
      if (item.closed == true &&
          item.type == JsonToken.OBJECTSTART &&
          item.child == null) {
        List<JsonToken> stack = new List<JsonToken>();
        JsonToken start = item;
        stack.add(start);
        for (int j = i + 1; j < len; j++) {
          int key = keys[j];
          JsonToken item = JsonToken.queue[key];
          if (item.type == JsonToken.OBJECTEND &&
              item.parnterId == start.parnterId) {
            item.child = null;
            item.partner.child = null;
            stack.add(item);
            if (stack.length > 2) {
              //send data
              List obj = this._object(stack);

              if (this._pipe != null) {
                this._pipe(obj[0], obj[1]);
              }
              for (String k in this._filters.keys) {
                if (new RegExp(k).hasMatch(stack[1].key)) {
                  this._filters[k](obj[0], obj[1]);
                }
              }
              if (this._trailing != null) {
                List<int> trailKeys = trailing.keys.toList();
                int k = trailKeys[trailKeys.length - 1];
                if (trailing[k]) {
                  if (trailing[k].runtimeType == bool) {
                    this._trailing(obj[0], obj[1]);
                  }
                }
              }
            }
            i = j;
            continue first;
          } else {
            int next = i + 1;
            deletes.add(key);
            deletes.add(keys[next]);
            stack.add(item);
          }
        }
      } else if (item.type == JsonToken.VALUE && item.closed == true) {
        if (item.value == 'true') {
          item.value = true;
        } else if (item.value == 'false') {
          item.value = false;
        } else if (item.value == 'null') {
          item.value = null;
        }
        RegExp r = new RegExp(r'^"|"$');
        String _data = item.value.replaceAll(r, "");
        if (this._pipe != null) {
          //send data
          this._pipe(_data, item.key);
        }
        for (String k in this._filters.keys) {
          if (new RegExp(k).hasMatch(item.key)) {
            this._filters[k](_data, item.key);
          }
        }
        if (this._trailing != null) {
          for (int k in trailing.keys) {
            if (trailing[k]) {
              if (trailing[k].runtimeType == bool) {
                this._trailing(_data, item.key);
              }
            }
          }
        }
        int next = i + 1;
        deletes.add(keys[next]);
        deletes.add(key);
      }
      if (trailing.keys.length > 0 &&
          (item.type == JsonToken.ARRAYSEPERATOR ||
              item.type == JsonToken.OBJECTEND) &&
          i > 2) {
        int _key_1 = keys[i - 3];
        int _key_2 = keys[i - 2];
        int _key_3 = keys[i - 1];
        JsonToken _item_1 = JsonToken.queue[_key_1];
        JsonToken _item_3 = JsonToken.queue[_key_3];
        if (_item_1.type == JsonToken.KEY) {
          deletes.add(_key_1);
        }
        if (_item_3.type == JsonToken.VALUEPAIR) {
          deletes.add(_key_2);
          deletes.add(_key_3);
        }
        deletes.add(key);
        var obj = {};
        RegExp r = new RegExp(r'^"|"$');
        obj[_item_1.value] = _item_3.value.replaceAll(r, "");
        if (_item_1.type == JsonToken.KEY &&
            _item_3.type == JsonToken.VALUEPAIR) {
          obj[_item_1.value] = _item_3.value.replaceAll(r, "");
          if (this._trailing != null) {
            this._trailing(obj, _item_3.key);
          }
          for (String k in this._filters.keys) {
            if (new RegExp(k).hasMatch(item.key)) {
              this._filters[k](obj, item.key);
            }
          }
        }
      }
    }
    for (int m = 0; m < len; m++) {
      int key = keys[m];
      JsonToken item = JsonToken.queue[key];
      if (item == null) {
        continue;
      }
      if (item.parent != null) {
        if (item.parent.closed) {
          item.closed = true;
          deletes.add(key);
        } else if (item.closed == true) {
          deletes.add(key);
        }
      } else if (item.closed == true) {
        deletes.add(key);
      }
      if (item.type == JsonToken.OBJECTSEPERATOR && !item.closed) {
        if ((m - 1) < len) {
          int previous = keys[m - 1];
          JsonToken sub = JsonToken.queue[previous];
          if (sub.type != JsonToken.KEY) {
            deletes.add(key);
          }
        }
      }
      if (item.type == JsonToken.KEY) {
        int next = m + 1;
        int nextnext = next + 1;
        if (nextnext < len) {
          JsonToken sub = JsonToken.queue[keys[nextnext]];
          if (sub.closed == true) {
            deletes.add(key);
          }
        }
      }
    }
    deletes.forEach((v) {
      if (JsonToken.queue.containsKey(v)) {
        JsonToken.queue.remove(v);
      }
    });
  }

  Reader pipe(Function func(dynamic value, String key)) {
    this._pipe = func;
    return this;
  }

  Reader trailing(Function func(dynamic value, String key)) {
    this._trailing = func;
    return this;
  }

  Reader done(Function func) {
    this._done = func;
    return this;
  }

  Reader fail(Function func(dynamic err)) {
    this._fail = func;
    return this;
  }

  Reader filter(String expression, Function func(dynamic value, String key)) {
    this._filters[new JsonExpression(expression).toPath()] = func;
    return this;
  }

  List<dynamic> _object(List<JsonToken> arr) {
    String string = '';
    String key = arr[1].key;
    arr.forEach((JsonToken d) {
      string += d.value;
    });
    return [JSON.jsonDecode(string), key];
  }
}
